/****************************************************************************
 File Name  : contour.c
 Purpose    : provides basic contour manipulation routines and serves as 
	      internal energy input for g-snake
 Release    : Version 1.0
 Date	    : Aug 31,1995

GSNAKE API is jointly developed by the Information Technology Institute (ITI), Singapore, and the School of Applied Science, Nanyang Technological
University (NTU), Singapore. 

These software programs are available to the user without any license or royalty fees. Permission is hereby granted to use, copy, modify, and distribute this software and its documentation for any purpose. ITI and NTU gives no warranty, express, implied, or statuary for the software and/or documentation provided, including, without limitation, waranty of merchantibility and warranty of fitness for a particular purpose. The software provided hereunder is on an "as is"  basis, and ITI and NTU has no obligation to provide maintenance, support, updates, enhancements, or modifications.

GSNAKE API is available for any UNIX compatible system. User feedback, bugs, or software and manual suggestions should be sent via electronic mail to one of the following, who may or may not act on them as he/she desires :

		asschan@ntu.ac.sg
		kflai@iti.gov.sg
***************************************************************************/

#include "gsnake.h"
#include "random.h"
#include "xwindow.h"

CONTOUR::CONTOUR( void ) 
{
	head = NULL;
	tail = NULL;
	numsnaxel = 0;
	sig_nu_sqr = 1;
	Z = 1 ;
	direction = 0 ;
}

CONTOUR::~CONTOUR( void )
{
	reset();
}

/*
 * initialize snaxels ( link list )
 */
int CONTOUR::initSnaxel( short numpts )
{
        register short i;
	SNAXEL *sxptr;

	/* delete all snaxel if exists */
	for ( sxptr = head; sxptr; sxptr = sxptr->getNext() )
		delete sxptr;

        /* number of snaxel */
        numsnaxel = numpts;

        if ( !(sxptr =  new SNAXEL) )
                return MEMORYERROR;

        sxptr->putPrev( NULL );

        /* set the forward and backward pointers */
        for(i=1, head=sxptr; i<numpts; i++) {

		SNAXEL *next;

		if( !(next = new SNAXEL) )
			return MEMORYERROR;

                sxptr->putNext( next );
                sxptr->getNext()->putPrev( sxptr );
                sxptr = sxptr->getNext();
                sxptr->putNext( NULL );
        }

        tail = sxptr; /* tail of snake */

        return NOERROR;
}

/*
 * _iMemFree all snaxel 
 */
void CONTOUR::reset( void )
{
	SNAXEL *sxptr, *next;

	for ( sxptr = head; sxptr; sxptr = next) {
		next = sxptr->getNext() ;
		delete sxptr ;
	}
		
	head = tail = NULL;
}

/*
 * init a round (closed) snake
 */
int CONTOUR::init( short _cg_row, short _cg_col, double radius, short numpts )
{
        double step, angle;
        SNAXEL *sxptr;
        int Imgrow = 2*_cg_row;
        int Imgcol = 2*_cg_col;
	int i=0;

        /* set gsnake parameters */
        mode = _CLOSED;
        numsnaxel = numpts;
	cg_row = _cg_row;
	cg_col = _cg_col;

        /* generate snaxels */
        if ( initSnaxel(numpts) == MEMORYERROR ) {
           fprintf(stderr," <CONTOUR::init> : memory allocation error\n");
           return MEMORYERROR;
        }

        step = 2*M_PI / (double) numpts;

        /* defaulte values for shape matrix */
        for (angle=0,sxptr=head; sxptr; angle+=step,sxptr=sxptr->getNext()) {

                sxptr->putCol( CAP(0,_cg_col+radius*cos(angle),Imgcol) );
                sxptr->putRow( CAP(0,_cg_row+radius*sin(angle),Imgrow) );
                sxptr->putAlpha( 0.5 / cos(2*M_PI/(double) numpts) );
                sxptr->putBeta( 0.5 / cos(2*M_PI/(double) numpts) );
                sxptr->putLambda( 0.99 );
        }

        return NOERROR;
}

/*
 * init a line (opened) snake
 */
int CONTOUR::init( short sx, short sy,     /* starting points */
                   short ex, short ey,     /* end points */
                   short numpts )           /* # of snaxels on this snake */
{
        double len_x, len_y;
        double dx, dy;
        register short i;
        SNAXEL *sxptr;

        /* calculate number of point needed  */
        len_x = (double) (ex - sx) ;
        len_y = (double) (ey - sy) ;

        /* set gsnake parameters */
        mode = _OPENED;
        numsnaxel = numpts;

        /* generate snaxels */
        if ( initSnaxel(numpts) == MEMORYERROR ) {
           fprintf(stderr," <CONTOUR::init> : memory allocation error\n");
           return MEMORYERROR;
        }

        dx = len_x/(double) (numsnaxel -1);
        dy = len_y/(double) (numsnaxel -1);

        for (i=0,sxptr=head; sxptr; i++,sxptr=sxptr->getNext()) {

                sxptr->putCol( dx * i + sx );
                sxptr->putRow( dy * i + sy );
                sxptr->putAlpha( 0.5 );
		sxptr->putBeta( 0.5 );
                sxptr->putLambda( 0.99 );
        }

        /* adjust head and tail of snake */
        adjust();

        return NOERROR;
}

/*
 * init snake by drag / click mouse
 */
int CONTOUR::init( IMAGE *ximg,        /* reference image */
                unsigned char blowup,   /* blowup ratio */
                INITMODE  theINIT,      /* click/drag mode */
                SNAKEMODE _mode,
                int spacing )       	/* desired spacing of snake */
{
        short numpts;
        short *xdata, *ydata;

        /* allocate space for coordinates */
        ydata = (short *)_iMemAlloc( MAX_SNAKEPTS * sizeof(short) );
        xdata = (short *)_iMemAlloc( MAX_SNAKEPTS * sizeof(short) );

        if ( !ydata || !xdata ) {
           fprintf(stderr, "<CONTOUR::init> : memory allocation error\n");
                return MEMORYERROR;
        }

        /* show reference image */
        ximg->show(blowup);
        ximg->show(blowup);

        /* read mouse input */
        ( theINIT == _CLICKMOUSE ) ?
                xwin_InitSnakeClick(ydata, xdata, &numpts) :
                xwin_InitSnakeDrag(ydata, xdata, &numpts, 2*spacing);

        /* set gsnake parameters */
        mode = _mode;

        /* init snaxels */
        if ( init( ydata, xdata, numpts, blowup) == MEMORYERROR ) {
           fprintf(stderr, "<CONTOUR::init> : memory allocation error\n");
           return MEMORYERROR;
        }

        /* compute average power of snaxels */
        computeAvgLength();

        /* learn shape matrix */
        computeShape();

        /* _iMemFree memory */
        _iMemFree ( ydata );
        _iMemFree ( xdata );

        return NOERROR;
}

/*
 * init gsnake by fill in snaxel value
 */
int CONTOUR::init( short *ydata, 	/* row data */
                   short *xdata,   	/* col data */
                   short numpts,   	/* number of snaxel */
                   unsigned char ratio)    /* blow up ratio */
{
        short register i;
        SNAXEL *sxptr;

	if ( numpts != numsnaxel )
		reset();

        /* generate snaxels */
	if ( !getHead() )
          if ( initSnaxel(numpts) == MEMORYERROR ) {
             fprintf(stderr," <CONTOUR::init> : memory allocation error\n");
             return MEMORYERROR;
          }

        for ( i=0, sxptr = getHead(); sxptr; i++, sxptr = sxptr->getNext() ) {
                sxptr->putRow( *(ydata + i) / ratio );
                sxptr->putCol( *(xdata + i) / ratio );
                /* sxptr->putLambda( 0.0 ); */
        }

        return NOERROR;
}

/*
 * init gsnake by fill in snaxel value
 */
int CONTOUR::init( double *ydata,		/* row data */
                   double *xdata,  		/* col data */
                   short numpts,		/* number of snaxel */
                   unsigned char ratio )	/* blow up ratio */
{
        short register i;
        SNAXEL *sxptr;

	if ( numpts != numsnaxel )
		reset();

        /* generate snaxels */
	if ( !getHead() )
          if ( initSnaxel(numpts) == MEMORYERROR ) {
             fprintf(stderr," <CONTOUR::init> : memory allocation error\n");
             return MEMORYERROR;
          }

        for ( i=0, sxptr = getHead(); sxptr; i++, sxptr = sxptr->getNext() ) {
                sxptr->putRow( *(ydata + i) / ratio );
                sxptr->putCol( *(xdata + i) / ratio );
                /* sxptr->putLambda( 0.0 ); */
        }

        return NOERROR;
}

/*
 * read from file
 */
int CONTOUR::read(char *filename)
{
        FILE *fid;      /* file identifier */
        SNAXEL *sxptr;  /* snaxel */
        int magic;      /* magic number */
        int _numsnaxel;
	float _sig_nu_sqr;
        SNAKEMODE _mode;
        float _col, _row, _alpha, _beta, _lambda, _z;
        short i;

        /* open file */
        if ( !( fid = fopen( filename, "r" ) ) ) {
           fprintf(stderr, "<CONTOUR :: read> : %s cannot be read\n", filename);
           return FILEIOERROR;
        }

        /* read magic number */
        fscanf(fid, "%x\n", &magic);

        if ( magic != CONTOURMAGICNUMBER ) {
            fprintf(stderr, "<CONTOUR :: read> : %s wrong format \n", filename);
            return FILEIOERROR;
        }

        /* read gsnake mode */
        fscanf(fid,"%d\n", &_mode);

        /* read number of snaxels */
        fscanf(fid,"%d\n", &_numsnaxel);

	/* read white noise var */
        fscanf(fid,"%f\n", &_sig_nu_sqr);

        /* init snaxel */
        mode =  _mode ;
        numsnaxel =  _numsnaxel ;
	sig_nu_sqr = (double) _sig_nu_sqr;
        if ( initSnaxel(_numsnaxel) == MEMORYERROR ) {
           fprintf(stderr," <CONTOUR::read> : memory allocation error\n");
           return MEMORYERROR;
        }

        /* read snaxel parameters */
        for ( sxptr = getHead(), i = 0; sxptr && i < _numsnaxel;
                sxptr = sxptr->getNext(), i++ ) {

                fscanf(fid,"%f ", &_col);
                fscanf(fid,"%f ", &_row);
                fscanf(fid,"%f ", &_alpha);
                fscanf(fid,"%f ", &_beta);
                fscanf(fid,"%f \n", &_lambda);

                sxptr->putCol( (double)_col );
                sxptr->putRow( (double)_row );
                sxptr->putAlpha( (double)_alpha );
                sxptr->putBeta( (double)_beta );
                sxptr->putLambda( (double)_lambda );
        }
        if ( fscanf(fid,"%f\n", &_z) > 0 )
		Z = _z ;
	else	Z = 1.0 ;

        /* close file */
        fclose(fid);
        
        computeAvgLength() ;
        
        return NOERROR;
}

/*
 * write to file
 */
int CONTOUR::write(char *filename)
{
        FILE *fid;      /* file identifier */
        SNAXEL *sxptr;  /* snaxel */

        /* open file */
        if ( !( fid = fopen( filename, "w" ) ) ) {
                fprintf(stderr, "<CONTOUR :: write> : %s cannot be write\n",
                        filename);
                return FILEIOERROR;
        }

        /* write magic number */
        fprintf(fid,"%x\n", CONTOURMAGICNUMBER);

        /* write to file */
        fprintf(fid,"%d\n", getMode());
        fprintf(fid,"%d\n", getNumSnaxel());
	fprintf(fid,"%e\n", getSigNuSqr());

        for ( sxptr = getHead(); sxptr; sxptr = sxptr->getNext() ) {

                fprintf(fid,"%f ",sxptr->getCol());
                fprintf(fid,"%f ",sxptr->getRow());
                fprintf(fid,"%f ",sxptr->getAlpha());
                fprintf(fid,"%f ",sxptr->getBeta());
                fprintf(fid,"%f\n",sxptr->getLambda());
        }

	fprintf(fid,"%e\n", getZ());

        /* close file */
        fclose(fid);
        return NOERROR;
}

/*
 * energy : model constraint
 */
double CONTOUR::EInternal( SNAXEL *now )	/* current snaxel */
{
        double l_inf;           /* infinity norm */
        double meanrow, meancol;

        if ( fabs(now->getAlpha() ) < VERY_SMALL &&
             fabs(now->getBeta() )  < VERY_SMALL )
                return 1.0;

        /* compute shape coeff. (alpha & beta) with infinity norm */
        now->meanPosition(           /* mean position */
                mode, cg_row, cg_col,
                head, tail, &meanrow, &meancol);
        
        l_inf = MAX( fabs( meanrow - now->getRow() ),
                     fabs( meancol - now->getCol() ) );
        return( SQR( l_inf / avglen ) );
}

/*
 * learn A matrix coefficients for a snake
 * input is in U coordinate
 */
void CONTOUR::computeShape()
{
        SNAXEL *now, *prev, *next;
        double Xi, Yi, Xalpha, Yalpha, Xbeta, Ybeta, denom;

  
        computeAvgLength();

        contourCentered();         /* convert to U coordinate */
    
 
        for ( now=getHead(); now; now=now->getNext() ) {

                prev = now->getPrev( mode, tail );
                next = now->getNext( mode, head );

                /* compute shape coefficient (alpha & beta) */
                Xi = now->getCol() / avglen;
                Yi = now->getRow() / avglen;
                Xalpha = prev->getCol() / avglen;
                Yalpha = prev->getRow() / avglen;
                Xbeta  = next->getCol() / avglen;
                Ybeta  = next->getRow() / avglen;

                denom = Xalpha * Ybeta - Xbeta * Yalpha;

                if ( fabs(denom) < VERY_SMALL ) {
                        now->putAlpha( 0.5 ); 
			now->putBeta(  0.5 );
		}
                else {
                        now->putAlpha( (Xi*Ybeta - Yi*Xbeta) / denom );
                        now->putBeta(  (Yi*Xalpha - Xi*Yalpha) /denom );
                }

        }

    
        imageCentered();         /*  convert to V coordinate */

        if ( mode == _OPENED ) adjust();
}

/*
 * using shape matrix and the last 2 contour points, regenerate the entire
 * contour
 */
CONTOUR::regenerate()
{
	MATRIX Ar ;	/* reduced shape matrix */
	register int i, j ;
	int status ;
	
	if(numsnaxel < 3) return PARAMERROR ;	/* cannot regenerate */
	
	if( (status = Ar.init(numsnaxel-2, numsnaxel-2) ) != NOERROR ) 
	    return status ;
	
	SNAXEL *sptr ;
	for(i=0, sptr = getHead()->getNext(); 
			i<numsnaxel-2; i++, sptr=sptr->getNext() ) {
	
		Ar.put(i, i, -sptr->getAlpha()) ;
		if(i < numsnaxel-3) Ar.put(i, i+1, 1.0) ;
		if(i < numsnaxel-4) Ar.put(i, i+2, -sptr->getBeta() ) ;
	}
	
	/* regeneration algorithm */
	MATRIX *ArInv = Ar.inverse() ;
	SNAXEL *Un = getTail() ;
	SNAXEL *Un1 = Un->getPrev();
	SNAXEL Vn, Vn1;
	
	computeCG();
	contourCentered() ;
	
	Vn.putRow(Un1->getRow() - Un1->getBeta()*Un->getRow());
	Vn.putCol(Un1->getCol() - Un1->getBeta()*Un->getCol());
	Vn1.putRow(-Un1->getPrev()->getBeta()*Un1->getRow());
	Vn1.putCol(-Un1->getPrev()->getBeta()*Un1->getCol());

	for(i=0, sptr = getHead(); i<numsnaxel-2; i++, sptr=sptr->getNext() ) {
	
		sptr->putRow(-ArInv->get(i, numsnaxel-4)*Vn1.getRow() -
			     ArInv->get(i, numsnaxel-3)*Vn.getRow() ) ;
		sptr->putCol(-ArInv->get(i, numsnaxel-4)*Vn1.getCol() -
			     ArInv->get(i, numsnaxel-3)*Vn.getCol() ) ;
	}
	
	delete ArInv ;
	
	imageCentered() ;
	return NOERROR ;
} 


/*
 * convert a snake from V (image-centered coordinates) to U (object centered)
 */
void CONTOUR::contourCentered()
{
        SNAXEL *sxptr;
	short positive = 1;


	/* check that the co-ordinates are not already in U form */
	/* if in contour centered form, some co-ordinates will be negative */
	sxptr = head;
	
	while (sxptr && positive)
	{
		if (sxptr->getRow()<0) positive =0;
		sxptr=sxptr->getNext();
	};
	
	/* Co-ordinates are in V form */
	if (positive)
	{
        	for(sxptr=head; sxptr; sxptr=sxptr->getNext()) {
                	sxptr->putRow( sxptr->getRow() - cg_row );
                	sxptr->putCol( sxptr->getCol() - cg_col );
        	};
     	};
}

/*
 * convert a snake from U to V coordinate
 */
void CONTOUR::imageCentered()
{
        SNAXEL *sxptr;
	short negative=0;

	/* check that the co-ordinates are not already in V form */
	sxptr = head;
	
	while ( (sxptr) && (!negative) )
	{
		if (sxptr->getRow()<0) negative =1;
		sxptr=sxptr->getNext();
	};

   	if (negative)
   	{

        	for(sxptr=head; sxptr; sxptr=sxptr->getNext()) {
                	sxptr->putRow( cg_row + sxptr->getRow() );
                	sxptr->putCol( cg_col + sxptr->getCol() );
        	};
	};
}

/*
 * compute average power (length)  of snaxels
 */
double CONTOUR::computeAvgLength()
{
        double sumlen = 0;              /* total power */
        short numpoints;                /* number of snaxel points */
        SNAXEL *sxptr, *next;
        register short i;

        /* compute center of gravity */
        computeCG();

        numpoints = (mode == _OPENED) ? numsnaxel - 1 :
                                        numsnaxel;

        for ( i=0, sxptr=getHead(); i<numpoints;
                sxptr=sxptr->getNext(), i++ ) {

                next = sxptr->getNext( mode, head );

                sumlen += hypot( (double) (sxptr->getRow() - next->getRow()),
                                 (double) (sxptr->getCol() - next->getCol()) );
        }

        avglen = sumlen / (double) numpoints;

        return ( avglen );
}

/*
 * compute center of gravity
 */
void CONTOUR::computeCG()
{
        SNAXEL *sxptr;

        /* reset center of gravity */
        cg_row = 0;
        cg_col = 0;

        for(sxptr=getHead(); sxptr; sxptr=sxptr->getNext()) {
                cg_row += sxptr->getRow();
                cg_col += sxptr->getCol();
        }

        cg_row = cg_row / (double) (numsnaxel);
        cg_col = cg_col / (double) (numsnaxel);
}
/*
 * use Monte Carlo estimation to compute Z : the normalizing constant
 */
void CONTOUR::computeZ()
{
	SNAXEL *sptr, *tSptr ;
	double Esum  ;
	int numSample ;
	float M ;
	double oldZ ;
	
	
	CONTOUR *temp = CONTOUR::duplicate() ;
	
	/* get average lambda */
	double avgLam = 0 ;
	for(sptr=getHead(); sptr; sptr=sptr->getNext())
		avgLam += sptr->getLambda() ;
	avgLam /= getNumSnaxel() ;	
	
	/* compute region for importance sampling and total size */
	M = 3*sqrt(sig_nu_sqr*(1.0/avgLam - 1.0))*computeAvgLength() ;
	
	Esum = 0 ;
	Z = 0 ;
	numSample = 1 ;
	do {
	     oldZ = Z ;
	
	     /* generate random assignment of snake */
	     for(sptr=getHead(), tSptr=temp->getHead() ; sptr && tSptr; 
	     		sptr=sptr->getNext(), tSptr=tSptr->getNext() ) {

	 	tSptr->putRow( sptr->getRow() + rand_uniform(-M, M) ) ;
	     	tSptr->putCol( sptr->getCol() + rand_uniform(-M, M) ) ;
	     }
	     
	     /* calculate internal energy */
	     double _Einternal = 0 ;
	     temp->computeAvgLength() ;
	     for(tSptr=temp->getHead(); tSptr; tSptr=tSptr->getNext())
	     	_Einternal += temp->EInternal(tSptr) / 
	     			(sig_nu_sqr*(1.0/tSptr->getLambda() - 1.0)) ;
	     
	     Esum += exp(-_Einternal) ;
	     
	     
	     Z =  Esum / numSample ;
   	     numSample++; 

   	} while( fabs(Z - oldZ)/Z > 0.001 || numSample < 100 ) ;
   	
    	Z =  ( pow((2*M+1), (double)(getNumSnaxel()) ) * Z) + 1.0 ;
      	
        delete temp ;
}



/*
 * expand snake
 * if spacing >  0 : desired spacing of expanded snake
 * if spacing <= 0 : desired number of snaxels inserted
 */
int CONTOUR::expand( double ratio,         /* ratio of expanding */
             	     int spacing)          /* spacing */
{

        SNAXEL *now, *next;
        double *ydata, *xdata;  /* snaxel coordinate */
        double dist;            /* dist. betw. snaxels */
        register short i, j;
        int numpts;             /* number of snaxels */
        int instNum;            /* # of pts inserted betw. snaxels */
        int counter;
        double dy, dx;

	if ( !spacing ) { /* if no new snaxels is inserted */

	    SNAXEL *sxptr;

	    for ( sxptr = getHead(); sxptr; sxptr = sxptr->getNext() ) {
                sxptr->putRow( sxptr->getRow() * ratio );
                sxptr->putCol( sxptr->getCol() * ratio );
	    }

	    return NOERROR;
        }

        /* allocate more memory to hold expanded locations */
        
	numpts = ROUNDOFF(ratio) * numsnaxel * ROUNDOFF( computeAvgLength() );
	
	
        ydata = (double *)_iMemAlloc( numpts * sizeof(double) );
        xdata = (double *)_iMemAlloc( numpts * sizeof(double) );

        if (!ydata || !xdata) {
           fprintf(stderr,"ERROR :<CONTOUR::expand> memory allocation error\n");
           return MEMORYERROR;
        }

        /* expand source snake and put the results into temporary buffer.
           The intersnaxel distance in the expanded snake is aprrox
           equal to "spacing" */

        now = getHead();
        ydata[0] = (ratio) * now->getRow();       /* starting point */
        xdata[0] = (ratio) * now->getCol();
        numpts = 1;
        counter = (mode==_CLOSED) ? numsnaxel + 1 : numsnaxel;

        for( j = 1; j < counter && now; j++, now = now->getNext() ) {

                /* find out how many points to be inserted between two
                   successive snaxels */

                next = now->getNext(mode, head);
                dy = (ratio) * ( next->getRow() - now->getRow() );
                dx = (ratio) * ( next->getCol() - now->getCol() );

                dist = MAX( fabs(dx), fabs(dy) );

                instNum = (spacing > 0) ?
                          ROUNDOFF( dist / (double) spacing ) - 1 :
                          -spacing;

                instNum = MIN( instNum, (int) (dist - 1) );

                for( i = 0 ; i < instNum  ; i++ ) {  /* insert the points */
                        ydata[numpts] = (ratio) * now->getRow() +
                                ROUNDOFF((i+1) * dy / (double) (instNum + 1));
                        xdata[numpts] = (ratio) * now->getCol() +
                                ROUNDOFF((i+1) * dx / (double) (instNum + 1));
                        numpts++ ;
                }

                /* put int the next point from scrc snake */
                if( j < numsnaxel ) {
                        ydata[numpts] = (ratio) * next->getRow() ;
                        xdata[numpts] = (ratio) * next->getCol() ;
                        numpts++;
                }
        }

        /* copy temporary buffer to expanded snake */
        if ( init(ydata, xdata, numpts) == MEMORYERROR )
                return MEMORYERROR;

        _iMemFree(ydata);            /* _iMemFree memory */
        _iMemFree(xdata);

        return NOERROR;
}


/*
 * copy snaxel value of snake
 */
void CONTOUR::copySnaxel(CONTOUR *target)
{
        SNAXEL *src;    /* source snaxel */
        SNAXEL *dest;   /* destination snaxel */

        src = getHead();                /* get head of src snake */
        dest = target->getHead();     /* get head of dest snake */

        /* start copying */
        for ( ; src; src = src->getNext(), dest = dest->getNext() ) {
                dest->putAlpha(	src->getAlpha() );
                dest->putBeta(	src->getBeta() );
		dest->putEint(	src->getEint() );
		dest->putEext(	src->getEext() );
                dest->putRow(	src->getRow() );
                dest->putCol(	src->getCol() );
                dest->putLambda(src->getLambda() );
                dest->putEsnaxel(src->getEsnaxel() );
        }
}


/*
 * Allows several functions of duplication
 	1. target NULL -> create new CONTOUR. All will be copied.
 	2. flag 1 -> copy snaxels only to target contour. For target NULL,
 	             all will be copied. This flag is ignored. 	
 	3. flag 0 -> copy all contour data. Will reset target contour and
 	             copy existing contour data
  */

CONTOUR *CONTOUR::duplicate(CONTOUR *target, short flag)
{
        CONTOUR *dupContour;

	/* If no existing contour create new one */
	if (!target)
	{

	   if ( !( dupContour = new CONTOUR ) ) {
               fprintf(stderr," <CONTOUR::duplicate> : memory allocation 	
               	               error\n");
               return NULL;
            }
	 }
	 else dupContour=target;   /* else assign pointer to target */


        /* If flag is 0, copy all data.
           Either creating a new contour or copying everything to target */ 

	if ( (!flag) || (!target) ) {
        	
        	dupContour->reset();
        	if ( dupContour->initSnaxel( numsnaxel ) == MEMORYERROR ) {
           	    fprintf(stderr,"\n<CONTOUR::duplicate> : memory allocation 
           	    `		  error\n");
     	 	    return NULL;
        	}
        	
        	/* copy parameter fields */
        	dupContour->putMode(mode);
        	dupContour->putNumSnaxel(numsnaxel);
        	dupContour->putCgRow(cg_row);
        	dupContour->putCgCol(cg_col);
        	dupContour->putDirection(direction); 
        	dupContour->putSigNuSqr(sig_nu_sqr);
        	dupContour->putZ(Z);            	       		
	}
	
              
        /* Check if target snake is the same as source snake. Otherwise
           unable to copy snaxels over. However, dupContour must still be
           returned, otherwise existing target snake will be lost */
           
        if ( dupContour->getNumSnaxel() == numsnaxel ) 
		copySnaxel( dupContour); 
	else {
	
	     printf("\nTarget snake and source snake have different number 
	                of snaxels. Did not copy.");
	}

   	dupContour->computeAvgLength();
 
 	return dupContour;
}



/*
 * adjust head and tail of gsnake
 */
void CONTOUR::adjust()
{
        SNAXEL *sxptr;

        /* adjust head */
        sxptr = head;
        sxptr->putBeta( 1.0/sxptr->getNext()->getAlpha() );
        sxptr->putAlpha( - (sxptr->getNext()->getBeta() / 
		sxptr->getNext()->getAlpha()) );

        /* adjust tail */
        sxptr = tail;
        sxptr->putAlpha( 1.0/sxptr->getPrev()->getBeta() );
        sxptr->putBeta(  - (sxptr->getPrev()->getAlpha() /
		sxptr->getPrev()->getBeta()) );
}

/*
 * affine invariant transform
 */
void CONTOUR::affineTransform( double sx,	/* x-scale changes */
				double sy,	/* y-scale changes */
				double rt,	/* rotation changes */
				double tx,	/* x-translation changes */
				double ty,	/* y-translation changes */
				double idx,	/* x-dilation changes */
				double idy )	/* y-dilation changes */
{
	SNAXEL *sxptr;
	double row, col;

	for( sxptr=getHead(); sxptr; sxptr=sxptr->getNext() ) {

		/* rotation and dilation */
		col = (cos(rt) + sin(rt)*idy)   * sxptr->getCol() +
                      (cos(rt)*idx + sin(rt))   * sxptr->getRow();
                row = ( -sin(rt)+ cos(rt)*idy ) * sxptr->getCol() +
                      ( -sin(rt)*idx + cos(rt)) * sxptr->getRow();

		/* scaling */
		col *= sx;
		row *= sy;

		/* translation */
		col += tx;
		row += ty;

		sxptr->putCol( col );
		sxptr->putRow( row );
	}
}

void CONTOUR::print()
{
	register short i;
        SNAXEL *sxptr;

	printf("numsnaxel = %d, mode = %s, avglen = %.3f\n", numsnaxel,
		( mode == _OPENED ) ? "opened" : "closed", avglen) ;
	printf("sig_nu_sqr = %.3e, Z = %.3f\n", sig_nu_sqr, Z) ;
	printf("COG : (%d %d)\n", ROUNDOFF(cg_row), ROUNDOFF(cg_col) ) ;
        for(i=0, sxptr=head; sxptr; i++, sxptr = sxptr->getNext())
                printf("(%d) col:%d row:%d\talpha:%f\tbelta:%f\tlambda:%f\n",
                        i, ROUNDOFF(sxptr->getCol()), 
			ROUNDOFF(sxptr->getRow()), 
			sxptr->getAlpha(), sxptr->getBeta(),
                        sxptr->getLambda());
	printf("\n");
}


void CONTOUR::show(IMAGE *backgnd, unsigned char blowup,
		   int pt_Xoffset, int pt_Yoffset, short expand)
{
	SNAXEL *sxptr;

	xwin_setexpandratio( expand, blowup );

        for ( sxptr = head; sxptr; sxptr=sxptr->getNext() ) {
            
                sxptr->show( backgnd, ROUNDOFF( sxptr->getRow() ), 
                	     ROUNDOFF( sxptr->getCol() ), pt_Xoffset, 
                	     pt_Yoffset ); 

        }
}


/* Method to display a contour on background. Contour will be shown
   centralised in the image */

void CONTOUR::display(short mag)
{
	SNAXEL *sptr;
	IMAGE background;
	int tx,ty;
	int minX, minY,maxX,maxY,length,height;
	tx=0;ty=0;
	short Row,Col;

	maxX=0;
	maxY=0;					
					
	minX=ROUNDOFF(head->getCol());
	minY=ROUNDOFF(head->getRow());		
	for (sptr=head; sptr;sptr=sptr->getNext())
	{
		tx=ROUNDOFF(sptr->getCol());
		ty=ROUNDOFF(sptr->getRow());
		if (tx<minX) minX = tx ;
			else if (tx>maxX) maxX = tx;
		if (ty<minY) minY = ty ;
			else if (ty>maxY) maxY = ty;	
	}
	
	length = maxX-minX;
	height = maxY-minY;
				
	/* Assign minimum image size */
	if (length<10) length= 30;
	if (height<10) height =30;	

	/* Setting all points with reference to minimum.  */
	for (sptr=head;sptr;sptr=sptr->getNext())
	{
		sptr->putRow(sptr->getRow()-minY+ROUNDOFF(height/2));
		sptr->putCol(sptr->getCol()-minX+ROUNDOFF(length/2));
	
	}
	
	background.init(height*2,length*2);
	for(ty=0;ty<background.getRow();ty++)
		for(tx=0;tx<background.getCol();tx++)
			background.put(ty,tx,20);
					
	background.show(mag);
	xwin_setexpandratio( 1, mag );

        for ( sptr = head; sptr; sptr=sptr->getNext() ) {

                Col = ROUNDOFF( sptr->getCol() ) ;
                Row = ROUNDOFF( sptr->getRow() ) ;
                
                
                sptr->show( &background, Row, Col );

        }

	for (sptr=head;sptr;sptr=sptr->getNext())
	{
		sptr->putRow(sptr->getRow()+minY-ROUNDOFF(height/2));
		sptr->putCol(sptr->getCol()+minX-ROUNDOFF(length/2));
	}

}
