/****************************************************************************
 File Name  : gsnake.c
 Purpose    : provides g-snake manipulation routines, includes energy
	      minimization, minimax regularization and initialization
	      by generalized Hough transform.
 Release    : Version 1.0
 Date	    : Aug 31,1995

GSNAKE API is jointly developed by the Information Technology Institute (ITI), Singapore, and the School of Applied Science, Nanyang Technological
University (NTU), Singapore. 

These software programs are available to the user without any license or royalty fees. Permission is hereby granted to use, copy, modify, and distribute this software and its documentation for any purpose. ITI and NTU gives no warranty, express, implied, or statuary for the software and/or documentation provided, including, without limitation, waranty of merchantibility and warranty of fitness for a particular purpose. The software provided hereunder is on an "as is"  basis, and ITI and NTU has no obligation to provide maintenance, support, updates, enhancements, or modifications.

GSNAKE API is available for any UNIX compatible system. User feedback, bugs, or software and manual suggestions should be sent via electronic mail to one of the following, who may or may not act on them as he/she desires :

		asschan@ntu.ac.sg
		kflai@iti.gov.sg
***************************************************************************/

#include "gsnake.h"
#include "xwindow.h"

GSNAKE::GSNAKE(EEXTTYPE etype, double glambda)
{
	EextType = etype;
	global_lambda = glambda;
	EdgeMap = NULL;
	GaussImg = NULL;
}

GSNAKE::~GSNAKE(void)
{
	reset();
}

void GSNAKE::reset(void)
{
	 EdgeMap = NULL;
	 GaussImg = NULL;
}

GSNAKE::genPyramid(short level, short verbose)
{
	if(EextType == _INTENSITY) 
	  return PYRAMID::generate(level, verbose, 1, 1, 1, 1, 1, _INTENSITY);
	else
	  return PYRAMID::generate(level, verbose) ;
}

/* localize gsnake at desired place. 
   19 May - For original version, localise will crash if called when pyramid 
            is not built.
            Implemented to check for this, will use root image if no pyramid.
            If no root image, will exit 
   
 */
int GSNAKE::localize(int _Qx, int _Qy, 
               int _NR, double _QR, 
               int _Nrxy, double _Qrxy,
               int _Nt,  double _Qt,
               int _Ndx, double _Qdx,
               int _Ndy, double _Qdy,
               double _R, double _THETA)	
{
	CONTOUR *contour;
	GHOUGH GHT_OBJ(_Qx, _Qy, _NR, _QR, _Nrxy, _Qrxy,
               	_Nt, _Qt, _Ndx, _Qdx, _Ndy, _Qdy, _R, _THETA);

        short level = PYRAMID::getLevel() - 1;
       
        if (level<0) {
        	printf("\nAuto generating pyramid to level 1 to localize.\n");
        	PYRAMID::generate(1,1);
        	level = 0;
        }
        
  	expand( LEVEL(-level) );
  	
  	GaussImg = getGauss(level) ;
	EdgeMap = getEdge(level) ;

 
  	/* generalize hough transform */
	switch(EextType) {
	    case _INTENSITY :
  		contour = GHT_OBJ.localize(this, GaussImg) ;
		break ;
	    case _EDGE :
		contour = GHT_OBJ.localize(this, EdgeMap) ;
		break ;
	    case _EDGEMAG :
		contour = GHT_OBJ.localize(this, EdgeMap->getMagImg());
		break ;
	}
        
	contour->duplicate( (CONTOUR *) this); 

	if ( _Qx > 1 || _Qy > 1 ) 
		fineLocalize( _Qx, _Qy); 

	GaussImg = getGauss(0) ;
        EdgeMap = getEdge(0) ;


	/* Expand contour to original size */
	CONTOUR::expand(LEVEL(level));
	
  
	delete contour ;

	return NOERROR;
}

/*
 *  perform fine correlation around center of gravity
    
    Edit : 22 June tmp now duplicates a contour instead of the whole gsnake
 
 */
 
void GSNAKE::fineLocalize(  int _qx,	/* translation resolution */
			    int _qy,
			    int _cg_col,/* centre of gravity */
			    int _cg_row )
{
	register short i, j ;
	SNAXEL *sxptr;
	int max_i, max_j ;
	double max_match, match ;

	/* at least correlate in 3x3 window */
	_qx = MAX(_qx, 3) ;
	_qy = MAX(_qy, 3) ;

	 /* get gaussian pyramid */
        GaussImg = PYRAMID::getGauss( 0 );
        EdgeMap  = PYRAMID::getEdge( 0 );

	/* move centre of gravity if neccessary */
	if ( _cg_col || _cg_row ) {
		translate( ROUNDOFF( _cg_col - cg_col ) , 
			   ROUNDOFF( _cg_row - cg_row ) );
	}
						    
	CONTOUR *tmp = CONTOUR::duplicate();
	
	max_i = max_j = 0 ;
	max_match = -getNumSnaxel() ;
	for (i = - (_qy/2) ; i <= _qy/2 ; i++)
	   for(j = - (_qx/2) ; j <= _qx/2 ; j++) {

		tmp->duplicate(this, 1);
		translate( j, i );
		match = 0;

		/* compute external energy */
		for (sxptr=getHead(); sxptr; sxptr=sxptr->getNext())
			match += EExternal( sxptr ) ;

		match = -match;	

		/* select maximun energy */
		if ( match > max_match ) {
			max_match = match;
			max_i = i;
			max_j = j;
		} 
	}

	tmp->duplicate(this, 1);
	translate( max_j, max_i );
	
	delete tmp;  
}

/*
 * Use a gaussian pyramid to perform coarse to fine minimization
 */
int  GSNAKE::minimize( 
		int segmentSpacing,   /* coarse search spacing */
                int numSearchSegment, /* region to move around */
                int snaxelSpacing,   /* snaxel spacing */
                unsigned char blowup, /* blow up ratio */
                int verbose,
                double theLambda,     /* lambda */
                int showImg,		/* show image ? */
		int img_Xoffset,	/* offset from image */
		int img_Yoffset)
{
        short curLevel;                 /* current gaussian pyramid level */
        int not_minimize = 1;		/* Initialise to not minimized */ 
	double old_lambda;
	
        /* get gaussian pyramid level */
        short level = PYRAMID::getLevel() - 1;

        if (level<0) {
        	printf("\nAuto generating pyramid to level 1 to minimize.\n");
        	PYRAMID::generate(1,1);
        	level = 0;
        }
        
        old_lambda = global_lambda;
        if( theLambda != _DEFINED_LAMBDA) 
             	global_lambda = theLambda;
        
	
	/* Reduce contour back to highest level */
	expand(LEVEL(-level)); 
       
        /* beginning from top-most level, do coarse to fine minimization */
        for (curLevel = level; curLevel >= 0; --curLevel) {

                not_minimize = mn_dymprog(curLevel,
                        segmentSpacing, numSearchSegment, blowup, verbose,
			showImg, img_Xoffset, img_Yoffset);

                 /* go down the levels if neccessary */
                if ( curLevel ) {

		        /* expand snake */
                        expand( LEVEL(1), snaxelSpacing );
                     
                        computeShape() ;

                        /* all verification step must use local minimax */
                        if(snaxelSpacing) global_lambda =  _LOCAL_MINMAX;
                }

        }
	global_lambda = old_lambda;

        return not_minimize;
}

/* 
 * marginalize gsnake to get probability 
 */
double GSNAKE::marginalize(int nhoodTangent, int nhoodNormal)
{
	LOCATIONS searchLocs ;
	nhoodTangent = 2*((int) nhoodTangent/2) + 1;
	nhoodNormal  = 2*((int) nhoodNormal/2)  + 1 ;
	
	/* must use lowest level */
        EdgeMap = getEdge(0) ;
        GaussImg = getGauss(0) ;
        
        if( searchLocs.init(numsnaxel, nhoodTangent*nhoodNormal) == MEMORYERROR)
        {
            fprintf(stderr, 
            	"ERROR : <GSNAKE::dymprog> memory allocation error\n");
	    return MEMORYERROR;
	}

	/* generate square window for marginalization */
	searchLocs.generateLoc(mode, head, tail, 
		nhoodNormal/2, nhoodTangent/2, 1, 0, 1) ;
	
	return dymprog(1, &searchLocs, 1) * 
			 exp(getNumSnaxel()/sig_nu_sqr) / getZ() ;		
}


/*
 * compute total energy of a gsnake object
 */
double GSNAKE::ESnake( short level, int verbose )    /* pyramid edge map */
{
        SNAXEL *pnow;           /* current snaxel */
        
        switch(EextType) {

            case _INTENSITY :
            	if( !(GaussImg = getGauss(level)))
			return PARAMERROR ;
            		break ;

            case _EDGE :
            case _EDGEMAG :
        	if ( !(EdgeMap = getEdge(level))  || 
		     !(GaussImg = getGauss(level)) )
			return PARAMERROR ;
			break;
       }
	
        /* compute average distant of snaxel coordinate */
        computeAvgLength() ;

        /* get initial value */
        Esnake = 0;     /* snake energy */
        Eext = 0;       /* average gradient power on snake */
    
        /* compute snake energy */

        for ( pnow = getHead(); pnow; pnow = pnow->getNext() ) {

                Esnake += ESnaxel( pnow,  _TRUE, verbose );
                Eext += pnow->getEext();                /* gradient power */
        }

        /* compute data strength information */
        Eext    /= (double) numsnaxel;  /* ave. gradient power */
  
        /* compute ave. snake energy */
        Esnake /= (double) numsnaxel;

	
        return Esnake;
}

/*
 * compute  total energy at current snaxel
 */
double GSNAKE::ESnaxel( SNAXEL *now,	/* current snaxel */
                        BOOLEAN store,
                        int verbose )
{
        double Eimage;  /* edge energy */
        double Emodel;  /* model energy */
        double Epoint;  /* total energy at current snaxel */

	double row = now->getRow();
	double col = now->getCol();

        /* check out of range */
        if (row != CAP(0, row, GaussImg->getRow()-1) ||
            col != CAP(0, col, GaussImg->getCol()-1) ) {

                if ( store == _TRUE ) {
                        now->putEext( 1 );
                        now->putEsnaxel( 1 );
                }
                return 1.0;
        }
        
        Eimage = EExternal(now);

        /* compute internal (model) energy */
        Emodel = EInternal( now );

        /* regularize */
        regularize( now, &Emodel, &Eimage, &Epoint);

        /* store energy term if neccessary */
        if (store == _TRUE) {
                now->putEint( Emodel );
                now->putEext( Eimage );
                now->putEsnaxel( Epoint );
        }

        if (verbose){
        	
           fprintf(stdout, "[%d, %d]\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n",
                 ROUNDOFF(col), ROUNDOFF(row), Emodel, Eimage, Epoint ,
		 now->getLambda() );
        }

        return Epoint;
}

/*
 * compute current snaxel energy based on 3 points
 */
double GSNAKE::ESnaxel( SNAXEL *now,	 	/* current snaxel */
                        int pt_ind,       	/* point index */
                        int loc_ind,      	/* location index */
                        int spacing,      	/* desired spacing */
                        LOCATIONS *searchLocs,  /* search locations */
                        int verbose )
{
	/* gaussian edge and image */
        register short i;
        double Emodel;          /* internal energy */
        double Eimage;          /* external energy */
        double Epoint;          /* snaxel energy */

        /* check out of range */
        if (now->getRow() != CAP(0, now->getRow(), GaussImg->getRow()-1) ||
            now->getCol() != CAP(0, now->getCol(), GaussImg->getCol()-1) )
            return 1.0;

        Emodel = EInternal( now );

        double MinEImage = 1.0;

        for (i=-spacing/2; i<=spacing/2; i++) {

            now->putRow( searchLocs->getRow( pt_ind, loc_ind ) + 
			 i*searchLocs->getNormalY( pt_ind ) );
            now->putCol( searchLocs->getCol( pt_ind, loc_ind ) + 
			 i*searchLocs->getNormalX( pt_ind ) );

	    Eimage = EExternal(now);
	   
           MinEImage = MIN( Eimage, MinEImage );

        }

        now->putCol( searchLocs->getCol( pt_ind, loc_ind ) );
        now->putRow( searchLocs->getRow( pt_ind, loc_ind ) );
        Eimage = MinEImage;

        /*regularize*/
        regularize( now, &Emodel, &Eimage, &Epoint );

        return Epoint;
}

/*
 * energy : edge force
 */
double GSNAKE::EEdge( SNAXEL *now,         /* current snaxel */
		      EDGE *edgeMap )     /* pyramid edge map */
{
        double mag;             /* edge magnitude */
        double Amodel;          /* gsnake object angle */
        double Adata;           /* edge map angle */
        double len_W;
        int mulp = 1;
        double uv_x, uv_y, uv0_x, uv0_y, uv1_x, uv1_y;

	short row,col;

	

	 /* check out of range */
        if (now->getRow() != CAP(0, now->getRow(), edgeMap->getRow()-1) ||
            now->getCol() != CAP(0, now->getCol(), edgeMap->getCol()-1) )       
            return 1.0;
   
        	
        /* get edge magnitude */
        mag = edgeMap->getMag( ROUNDOFF( now->getRow() ),
                               ROUNDOFF( now->getCol() ) );

        if ( mag < VERY_SMALL)
                return 1.0;

        /* return edge magnitude if opened snake and at first or last pt */
        if ( mode == _OPENED &&
             ( now->getPrev() == NULL || now->getNext() == NULL ) )
        	return 1 - mag;

        /* compute snaxel tangent direction */
        now->getDirectionVec( mode, head, tail, &uv0_x, &uv0_y, &uv1_x, &uv1_y);

        /* tangent vector at current snaxel */
        uv_x = uv0_x + uv1_x;
        uv_y = uv0_y + uv1_y;

        len_W = hypot( uv_x, uv_y );
        if ( len_W < VERY_SMALL )
                return 1.0;

        /* compute image energy */
        Amodel = ATAN2( uv_x, -1*uv_y );
        Adata = edgeMap->getAng( ROUNDOFF( now->getRow() ),
                                 ROUNDOFF( now->getCol() ) );

        if ( direction )
                mulp = EDGEDIR(-1*uv_y, uv_x, cos(Adata), sin(Adata)) *
                        direction;

        return (        /* compute projection angle */
                1 - mulp*mag*fabs(cos(Amodel-Adata)) );
}

double GSNAKE::EMag( SNAXEL *now,         /* current snaxel */
                     EDGE *edgeMap )     /* pyramid edge map */
{
	

	 /* check out of range */
        if (now->getRow() != CAP(0, now->getRow(), edgeMap->getRow()-1) ||
            now->getCol() != CAP(0, now->getCol(), edgeMap->getCol()-1) )
            return 1.0;

	short row = ROUNDOFF( now->getRow() );
	short col = ROUNDOFF( now->getCol() ); 

	    
	return ( MAXMAG - edgeMap->getMag(row, col) );
}

double GSNAKE::EInten( SNAXEL *now,		/* current snaxel */
		       IMAGE *gaussImg )	/* gaussian image */
{
	 /* check out of range */
        if (now->getRow() != CAP(0, now->getRow(), gaussImg->getRow()-1) ||
            now->getCol() != CAP(0, now->getCol(), gaussImg->getCol()-1) )
            return 1.0;

	short row = ROUNDOFF( now->getRow() );
	short col = ROUNDOFF( now->getCol() ); 


	return ( MAXMAG - gaussImg->get(row, col) );
}

double GSNAKE::EInternal(SNAXEL *now)
{
	if (now)
	{
		Eint = CONTOUR::EInternal(now);
		return Eint;
	
	} else return PARAMERROR;
}


/* Top level external energy method. Will return PARAMERROR if null
   oassed in. */
double GSNAKE::EExternal(SNAXEL *now)
{  
	double Eimage;

     	if (now) {	
     		
		switch( EextType ) {

	   		case _EDGE :
		   	 	Eimage = EEdge(now, EdgeMap);
             	   		break;

	   		case _EDGEMAG :
				Eimage = EMag( now, EdgeMap);			
                		break;

	   		case _INTENSITY :
				Eimage = EInten(now, GaussImg);
				break;
	   	}
	   	
	   	return Eimage;

	} else 
		return PARAMERROR;	
}


/*
 * regularizetion : minmax principle or global/local lambda
 */
void GSNAKE::regularize( SNAXEL *now,	/* current snaxel */
			     double *Emodel,    /* model energy */
                             double *Eimage,    /* image energy */
                             double *Epoint )   /* snaxel energy */
{
        if ( global_lambda == _LOCAL_MINMAX )   /* local minimax */
                *Epoint = MAX(*Emodel, *Eimage) + 0.0001*MIN(*Emodel, *Eimage);

        else if ( global_lambda == _LOCAL_LAMBDA ) {
                *Emodel *= now->getLambda() / ( 1.0 - now->getLambda() );
                *Epoint = *Emodel + *Eimage;
           }

        /* combine Emodel and Eedge by global model parameter */
	else if (( global_lambda < 1.0 ) && (global_lambda > 0.0) )
                *Epoint = global_lambda / ( 1.0 - global_lambda ) * 
			*Emodel + *Eimage;
	else
		*Epoint = 1;
}


/* Performs duplication of snake. If target exists, will copy contour and
   parameters. If target does not exist, will create target, copy pyramid 
   and contours and parameters. 
     */
   
GSNAKE *GSNAKE::duplicate(GSNAKE* target)
{
	GSNAKE *dupSnake;
	
	/* If no existing snake create new one */
	if (!target){
	   if ( !( dupSnake = new GSNAKE ) ) {
               fprintf(stderr," <GSNAKE::duplicate> : memory allocation 	
               	               error\n");
               return NULL;
            }
	    PYRAMID::duplicate(dupSnake);
	 }
	 else dupSnake = target;   /* else assign pointer to target */

	CONTOUR::duplicate(dupSnake);
		
	/* Copy GSNAKE parameters */
	dupSnake->global_lambda = global_lambda;
	dupSnake->EextType      = EextType;
 	dupSnake->Eext          = Eext;
 	dupSnake->Esnake        = Esnake; 

	return dupSnake;
}


/*
 * deform contour by mouse clicking
 */
void GSNAKE::deform(unsigned char blowup, short expand)
{
	register short i;
	SNAXEL *sxptr;
		   
	double xdata[numsnaxel];
	double ydata[numsnaxel];
	
			    
	for ( i=0, sxptr=getHead(); i<numsnaxel && sxptr;
		i++, sxptr=sxptr->getNext() ) {

		xdata[i] = sxptr->getCol()*blowup;
		ydata[i] = sxptr->getRow()*blowup;
	}

	show(blowup,expand,1);								       
	xwin_ShapeContour( rawImg->ximg, ydata, xdata, numsnaxel );
									
	CONTOUR::init( ydata, xdata, numsnaxel, blowup); 
}

/*
 * draw snake with given offset
 */
void GSNAKE::showLine( unsigned char blowup,	/* blow up ratio */
		   int Xoffset,			/* x offset */
		   int Yoffset )		/* y offset */
{
        SNAXEL *sxptr;
	short nowrow, nowcol;
	short prevrow, prevcol;
	short register i;

	prevcol = ROUNDOFF( getHead()->getCol() ) + Xoffset;
	prevrow = ROUNDOFF( getHead()->getRow() ) + Yoffset;

        for ( i = ( mode == _CLOSED ? 0 : 1 ),
		sxptr = getHead()->getNext(); ( i < numsnaxel ) && sxptr; 
		i++, sxptr=sxptr->getNext(mode, head) ) {

                nowcol = ROUNDOFF( sxptr->getCol() ) + Xoffset;
                nowrow = ROUNDOFF( sxptr->getRow() ) + Yoffset;

		xwin_DrawLine( prevcol, prevrow, nowcol, nowrow, blowup );

		prevcol = nowcol;
		prevrow = nowrow;

        }
}

/*
 * draw snake point
 */
void GSNAKE::show(  unsigned char blowup, 	/* blow up ratio */
		   short expand,                /* expand ratio */
               	   int showImg,			/* show image ? */
		   int img_Xoffset, 		/* image offset */
		   int img_Yoffset,
		   int pt_Xoffset,
		   int pt_Yoffset )
{
        SNAXEL *sxptr;
        
        if(!rawImg) {
          	fprintf(stderr, "GSNAKE does not contain a raw image.") ;
          	fprintf(stderr, " Using CONTOUR::display()\n") ;
          	CONTOUR::display(blowup) ;
        }
        
        else {

	 if ( showImg )
		rawImg->show( blowup, 1, img_Xoffset, img_Yoffset );

	 else 	
          for ( sxptr = getHead(); sxptr; sxptr=sxptr->getNext() )
		xwin_clrpoint( rawImg->ximg, 
			     (ROUNDOFF( sxptr->getCol() )+ img_Xoffset)*blowup, 
			     (ROUNDOFF( sxptr->getRow() )+ img_Yoffset)*blowup);


	 CONTOUR::show(rawImg,blowup, pt_Xoffset,pt_Yoffset, expand);
	}

}

/*
 * display new location of snaxel
 */
void GSNAKE::move(  GSNAKE *old_gsnake,	/* old gsnake */
		   int pt_Xoffset,	/* point offset */
		   int pt_Yoffset )
{
	SNAXEL *old_sxp;
	SNAXEL *new_sxp;

	for( new_sxp = getHead(), old_sxp = old_gsnake->getHead(); 
	     new_sxp && old_sxp; new_sxp = new_sxp->getNext(),
			 	 old_sxp = old_sxp->getNext() ) {

		short Col = ROUNDOFF( old_sxp->getCol() + pt_Xoffset) ;
                short Row = ROUNDOFF( old_sxp->getRow() + pt_Yoffset) ;

		new_sxp->show( rawImg, Row, Col, pt_Xoffset,pt_Yoffset );
	}
}
