/****************************************************************************
 File Name  : gsnake.h
 Purpose    : Header file for GSNAKE API
 Release    : Version 1.0
 Date	    : Aug 31,1995

GSNAKE API is jointly developed by the Information Technology Institute (ITI), Singapore, and the School of Applied Science, Nanyang Technological
University (NTU), Singapore. 

These software programs are available to the user without any license or royalty fees. Permission is hereby granted to use, copy, modify, and distribute this software and its documentation for any purpose. ITI and NTU gives no warranty, express, implied, or statuary for the software and/or documentation provided, including, without limitation, waranty of merchantibility and warranty of fitness for a particular purpose. The software provided hereunder is on an "as is"  basis, and ITI and NTU has no obligation to provide maintenance, support, updates, enhaucements, or modifications.

GSNAKE API is available for any UNIX compatible system. User feedback, bugs, or software and manual suggestions should be sent via electronic mail to one of the following, who may or may not act on them as he/she desires :

		asschan@ntu.ac.sg
		kflai@iti.gov.sg
****************************************************************************/


#ifndef GSNAKELIB
#define GSNAKELIB

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* error numbers */
#define NOERROR		 0 
#define FILEIOERROR	-1
#define MEMORYERROR	-2
#define PARAMERROR	-3

/* constants */
#define GRAYLEVELRANGE		255
#define CONTOURMAGICNUMBER	0x1010
#define VERY_SMALL		1e-10
#define VERY_BIG		1e+10
#define MAX_SNAKEPTS   		100
#define MAXMAG                  1.0

/* macros */
#define ROUNDOFF(x)		((int) ( (x > 0) ?  (x+0.5) : (x-0.5)))
#define LEVEL(ratio)		pow( 2.0, (double) ratio ) 
#define SIGN(x)			( ((int) x) ? ( (x > 0) ? 1 : -1 ) : 0 )
#define EDGEDIR(r1,c1,r2,c2)    ( SIGN(r1*r2 + c1*c2) )
#define MAX(x, y)		( (x > y) ? (x) : (y) ) 
#define MIN(x, y)		( (x < y) ? (x) : (y) )
#define CAP(low, x, upp)        ( (x < low) ? low : ( (x > upp) ? upp : x ) )
#define SQR(a)                  ( (a) * (a) )
#define RADIAN(angle)           ((double) angle/180.0 * M_PI )
#define ANGLE(radian)           (180.0 * radian / M_PI)
#define DEGREE(r)               ( r*180.0 / M_PI)
#define ATAN2(y, x)		((fabs(x) > VERY_SMALL)? atan2(y,x) : \
				 (y>0) ? 0.5*M_PI : -0.5*M_PI )

/* memory macros */
#define _iMemAlloc(x)		malloc(x)
#define _iMemRealloc(ptr, x)	( (ptr) ? realloc(ptr, x) : malloc(x) )
#define _iMemFree(x)		if(x) { free(x) ; x = NULL ; } 


#define _LOCAL_MINMAX   -1.0    /* flag at model for LMINMAX */
#define _LOCAL_LAMBDA    2.0    /* localize regularization parameter used */
#define _DEFINED_LAMBDA	-2.0	/* used previously defined lambda */

typedef enum _InputImageFormat { _bin, _ras };
typedef enum { _TRUE=1, _FALSE=0 } BOOLEAN ;

typedef enum {          /* type of energy used as image energy */
        _EDGE,          /* use edges : boundary detection */
        _EDGEMAG,       /* edge magnitude only : without direction info */
        _INTENSITY      /* use intensity. e.g. character detection */
} EEXTTYPE ;

typedef enum {		   /* methods classification */
	_MARGIN_PROB = 0,  /* summation of probablity for classification */
	_DEFORM_PROB = 1,  /* MAP probability */
	_DEFORM_MATCH = 2, /* match of deformable template */
	_RIGID_MATCH  = 3  /* match of rigid template */
} CLASSTYPE;

typedef enum {          /* open/connected snake */
	_OPENED = 0,    /* two end points are not connected */
	_CLOSED = 1     /* end points are connected */
} SNAKEMODE ;

typedef enum {
        _DRAGMOUSE,     /* manual initialization by dragging the mouse */
        _CLICKMOUSE,    /* manual initialization by clicking the mouse */
        _LOADTEMPLATE,  /* load templates from a given learned file */
	_INITTEMPLATE	/* init line or circle templates from program */
} INITMODE ;

/* Image object stores Image data in the form of a matrix of size row by col 
   data* points to start address of the matrix. ximg* points to start of 
   xwindow image, with magnifying factor magnify. */ 

class IMAGE {

    protected:
	float	*data;		/* image data */
	int	row, col;	/* row and col of image */

    public:
	XImage *ximg;
    
    protected:
	double innerProd (IMAGE *InputImg, int InpRowIndex, int InpColIndex);

    public :
	IMAGE();	
	~IMAGE(); 
	void reset();
	
	init(int _row, int _col);	
	void put(int m, int n, float val)	
		{ *( data+ (m)*col + (n) ) = val; } 
	float get(int m, int n)			
		{ return *( data+(m)*col+(n) ); }
	getRow() {return row;}	
	getCol() {return col;}
	void print();	
	void generateX(unsigned char blowup=1);	

	void show(unsigned char blowup=1,int num=1,
			   int h_offset=0, int v_offset=0);
	read (char *filename);   
	write(char *filename, short format=_bin);
	void condition(double low_pct=0.9, double high_pct=0.95, double
		       low_val=0.2, double high_val=0.9, double imm_pow=1.0); 	

	void initAsGauss();
	IMAGE *correlate (IMAGE *InputImg, int RowStep, int ColStep,
				char verbose=0);
	IMAGE *IMAGE::cut(int sx,int sy,int height, int length);
	int fill(float val, int sx=0,int sy=0, int length=0, int height=0);
	IMAGE *copy() {return cut(0,0,col,row);}			 
};

/* Provide advance matrix manipulation routines for IMAGE object. 
 */ 
class MATRIX : public IMAGE {
    public :
    	init(short _row, short _col, char identity=0);
    	swapRow(short i, short j );
    	void dump(char *header=NULL);
    	MATRIX * transpose (void) ;
    	MATRIX * operator+ (MATRIX& mtx) ;
    	MATRIX * operator- (MATRIX& mtx) ;
    	MATRIX * operator* (MATRIX& mtx) ;
    	MATRIX * inverse (void) ;
};

/*Edge object stores Magnitude and Angle information of the edge gradient
  in the form of two image objects. This way of representing the information
  serves to reduce coding needed for application */ 


class EDGE {

   protected :
	IMAGE *Mag;
	IMAGE *Ang;

   protected :
	init(int _row, int _col);
   	
   public :
	EDGE(){Mag=Ang=NULL;}
	~EDGE(){reset(); }
	void reset() ;
	float getMag(int m,int n) { return Mag->get(m,n); }
	float getAng(int m,int n) { return Ang->get(m,n); }
	void putMag(int m, int n, float val)
		{ Mag->put(m,n,val); }
	void putAng(int m, int n, float val)
		{ Ang->put(m,n,val); }
	void show(unsigned char magnify=1, int h_offset=0, int v_offset=0);
	compute(IMAGE *img, int verbose=1,
			double low_pct=0.9, double high_pct=0.95, double 	
		       low_val=0.2, double high_val=0.9, double imm_pow=1.0);
	getRow(int MagData=1) { return (MagData) ? Mag->getRow():Ang->getRow();}
	getCol(int MagData=1) { return (MagData) ? Mag->getCol():Ang->getCol();}
	
	void putMagImg( IMAGE *magImg ) { Mag = magImg; }
	void putAngImg( IMAGE *angImg ) { Mag = angImg; }
	IMAGE *getMagImg(void) { return Mag ; }
	IMAGE *getAngImg(void) { return Ang ; }
	
};	 		


class PYRAMID {

   protected :
   	short numLevel ;		/* number of level */
   	EDGE **edgemap ;		/* edge map at each level */
   	IMAGE **gaussImg ;		/* gaussian smoothed image */
   	
   protected :
	int init(short level);   
   
   public :
	IMAGE *rawImg;			/* raw image */

   	PYRAMID();
   	~PYRAMID();
	void reset();
   	generate(short level, int verbose=0,double low_pct=0.9, 	
		      double high_pct=0.95, double low_val=0.2,	
		      double high_val=0.9, double imm_pow=1.0,
		      EEXTTYPE = _EDGE);
	int duplicate( PYRAMID *pyramid );
   	EDGE *getEdge(short level_id) 
   		{ return ( (level_id>numLevel-1) || (level_id <0) ) ? 
			 NULL: edgemap[level_id]; }	
      	IMAGE *getGauss(short level_id)
      		{ return ( (level_id>numLevel-1) || (level_id <0) ) ? 
			 NULL: gaussImg[level_id]; }	
      	void putRawImg( IMAGE *img ) { rawImg = img->copy();}
	int putRawImg( char *filename );
	short getLevel(void) { return numLevel; }
	void print(int level);
	void show(unsigned magnify,int level=1);
};

class SNAXEL {

    protected :
	double row, col ;       /* snaxel coordinate */
	double alpha, beta ;    /* shape coefficients */
	double lambda ;         /* local regularization parameters */
	double Eint, Eext ;     /* internal/external energy */
	double Esnaxel; 	/* snaxel energy */
	SNAXEL *next;		/* next snaxel */
	SNAXEL *prev;		/* previous snaxel */

    public :

	SNAXEL();
	~SNAXEL() {;}
	double	getNormalAng( SNAKEMODE mode, SNAXEL *head, SNAXEL *tail);
	
	void	getDirectionVec( SNAKEMODE mode, 
		SNAXEL *head, SNAXEL *tail, 
		double *Uv0_x, double *Uv0_y,
		double *Uv1_x, double *Uv1_y);
	void	getNormalVec( SNAKEMODE mode,
		SNAXEL *head, SNAXEL *tail,
		double *nv_x, double *nv_y );

	void	meanPosition( SNAKEMODE mode, double cg_row,
		double cg_col, SNAXEL *head, SNAXEL *tail, 
		double *meanrow, double *meancol);

	/* interface */
	void	show( IMAGE *img, short old_Row, short old_Col,
	              int pt_Xoffset=0, int pt_Yoffset=0);

        double  getRow(void) {return row;}
	double  getCol(void) {return col;}
		 
	double  getAlpha(void) {return alpha;}
	double  getBeta(void) {return beta;}
	double  getLambda(void) {return lambda; }
	double  getEsnaxel(void) {return Esnaxel; }
	double  getEint(void) {return Eint; }
	double  getEext(void) {return Eext; }
	

	SNAXEL	*getNext(void) { return next; }
	SNAXEL	*getPrev(void) { return prev; }
	SNAXEL	*getNext(SNAKEMODE mode, SNAXEL *head);
	SNAXEL	*getPrev(SNAKEMODE mode, SNAXEL *tail);

        void  putRow( double _row ) {row = _row;}
	void  putCol( double _col ) {col = _col;}
	void  putPrev( SNAXEL *_prev ) {prev = _prev;}
	void  putNext( SNAXEL *_next ) {next = _next;}
		 
	void  putAlpha( double _alpha ) {alpha = _alpha;}
	void  putBeta( double _beta ) {beta = _beta;}
	void  putLambda( double _lambda ) {lambda = _lambda; }
	void  putEsnaxel( double _Esnaxel ) {Esnaxel = _Esnaxel;}
	void  putEint( double _Eint ) {Eint = _Eint; }
	void  putEext( double _Eext ) {Eext = _Eext; }
};


class CONTOUR {

    protected :
	SNAXEL *head;		/* head of snake */
	SNAXEL *tail;		/* tail of snake */
	SNAKEMODE mode;		/* opened or closed snake */
	short numsnaxel;	/* number of snaxel */
	double cg_row, cg_col;	/* center of gravity */
	double avglen;		/* ave. power (length) of snake */
        double sig_nu_sqr ;     /* white noise variance on gradient power */
        double Z ;		/* normalizing constant */
        short direction;        /* [0/-1/+1] preset gradient direction */

    protected :
	int initSnaxel( short numsnaxel );
	void adjust();


    protected :
	int init( short *ydata, short *xdata, 
		short numpts = 16, unsigned char ratio = 1 );
	int init( double *ydata, double *xdata, 
		short numpts = 16, unsigned char ratio = 1 );
	void copySnaxel( CONTOUR *target );


    public :

	CONTOUR(void);
	~CONTOUR(void);
	void reset(void);

	/* Internal Energy */
        double  EInternal( SNAXEL *sxptr );

	/* initialize contour routines */
	int init( short _cg_row, short _cg_col, 
		double radius, short numpts = 16);
	int init( short sx, short sy, short ex, short ey, 
		short numpts = 16 );
	int init( IMAGE *ximg,
		  unsigned char blowup = 1, 
		  INITMODE  theINIT = _CLICKMOUSE, 
		  SNAKEMODE _mode = _CLOSED,
		  int spacing = 5 );
	int read( char *filename );

	/* I/O routines */
	int write( char *filename );

	CONTOUR *duplicate(CONTOUR *target=NULL, short snx=0);

	int expand( double ratio, int spacing = 0 );
	/* transformation routines */
	void affineTransform( double sx, double sy, double rt, 
			  double tx, double ty, double idx, double idy );

	void rotate( double angle )
		{ affineTransform( 1, 1, RADIAN(angle), 0, 0, 0, 0 ); }
	void translate( int dx, int dy )
		{ affineTransform( 1, 1, 0, dx, dy, 0, 0 ); }
	void scale( double sx, double sy )
		{ affineTransform( sx, sy, 0, 0, 0, 0, 0 ); }
	void dilate(double idx, double idy)
		{ affineTransform(1,1,0,0,0,idx,idy); }

	
	/* compute contour shape */
	void computeShape();		/* learn shape parameters */
	void computeCG();		/* compute center of gravity */
	void computeZ() ;		/* compute norm. constant Z */
	double computeAvgLength(); 	/* compute average length */
	int regenerate();		/* regenerate contour from A */
	void contourCentered() ;	/* V to U coordinate */
	void imageCentered() ;		/* U to V coordinate */


	/* interface */
	void display(short mag=1);	
	void print();
	void show(IMAGE *backgnd, unsigned char blowup=1,
		   int pt_Xoffset=0, int pt_Yoffset=0,short expand=1);


	SNAXEL *getHead(void) { return head; }
	SNAXEL *getTail(void) { return tail; }

	SNAKEMODE getMode(void) { return mode; }
	short getNumSnaxel(void) { return numsnaxel; }
	double getSigNuSqr(void) { return sig_nu_sqr; }
	double getZ(void)     { return Z ; }
	double getCgRow(void) { return cg_row; }
	double getCgCol(void) { return cg_col; }

	void putCgRow(double _row) { cg_row = _row;}
	void putCgCol(double _col) { cg_col = _col;}
	void putMode(SNAKEMODE _mode) { mode = _mode; }
	void putNumSnaxel(short _numsnaxel) {numsnaxel = _numsnaxel; }
	void putSigNuSqr(double _sig_nu_sqr){ sig_nu_sqr = _sig_nu_sqr; }
	void putDirection(short _direction){ direction = _direction; }
	void putZ(double _Z) { Z = _Z ; }
};

class LOCATIONS {

    protected :
	short  numpts;	/* number of points */
	short  numlocs;	/* number of assigned locations */
	double **row;	/* possible location to move around */
	double **col;
	double **OldSum;/* for classification purpose */
	double **NewSum;
	double **OldMax;   /* Energy minimisation purpose */
	double **NewMax;
	short  ***MaxInd;  /* for backtracking purpose */ 
	short  *MI;        
	double *normalX; /* normal vector */
	double *normalY;

    public :
	LOCATIONS(void) {;}
	~LOCATIONS(void) ;
	int init(short _numpts, short _numlocs);
	void clearLoc();
	void generateLoc(
		SNAKEMODE mode, SNAXEL *head, SNAXEL *tail, 
		int NhoodX, int NhoodY, int spacing, 
		int rotate,short verbose =0 );

	/* get interface */
	short getNumLoc(void) { return numlocs; }

	double getRow( short i, short j ) { return row[i][j]; }
	double getCol( short i, short j ) { return col[i][j]; }

	double getOldSum( short i, short j ) { return OldSum[i][j]; }
	double getNewSum( short i, short j ) { return NewSum[i][j]; }

	double getOldMax( short i, short j ) { return OldMax[i][j]; }
	double getNewMax( short i, short j ) { return NewMax[i][j]; }

	double getNormalX( short i ) { return normalX[i]; }
	double getNormalY( short i ) { return normalY[i]; }

	short getMI( short i ) { return MI[i]; }
	short getMaxInd( short i, short j, short k ) 
				{ return MaxInd[i][j][k]; }

	/* put interface */
	void putRow( short i, short j, double val ) { row[i][j] = val; }
	void putCol( short i, short j, double val ) { col[i][j] = val; }

	void putOldSum( short i, short j, double val ) { OldSum[i][j] = val; }
	void putNewSum( short i, short j, double val ) { NewSum[i][j] = val; }

	void putOldMax( short i, short j, double val ) { OldMax[i][j] = val; }
	void putNewMax( short i, short j, double val ) { NewMax[i][j] = val; }

	void putnormalX( short i, double val ) { normalX[i] = val; }
	void putnormalY( short i, double val ) { normalY[i] = val; }

	void putMI( short i, short val ) { MI[i] = val; }
	void putMaxInd( short i, short j, short k, short val )
					 { MaxInd[i][j][k] = val; }
};

class GHOUGH {

    protected :
         IMAGE **Planes;		/* Planes of accumulator cells */
        CONTOUR **Template;      	/* Templates under transformations */
        double **Angle;         	/* desired gradient angles */
        int NumPlane;           	/* actual # of planes */
        int Qx, Qy;        	        /* X, Y resolutions of planes */
        double QR, Qrxy;        	/* scale resolutions */
        double Qt;              	/* theta resolutions */
        double Qdx, Qdy;    		/* dilation resolutions */
        int NR, Nrxy, Nt, Ndx, Ndy; 	/* # of cells in each of the axis */
	double R;			/* scaling factor */
	double THETA;			/* initial angle for ref. line angle */
	int plane_max;			/* store index for blanking */
	int row_max;
	int col_max;

    protected :
        
        int init(int row, int col);
        void transform(CONTOUR *reference, char comAngleFlag=1);
    
        void correlate(EDGE *edgeMap);
	void correlate(IMAGE *gaussImg);

       CONTOUR *getMax(void);

       reset(void) ;

    public :
        GHOUGH(int _Qx=1, int _Qy=1, 
               int _NR=1, double  _QR=0.25, 
               int _Nrxy=1, double _Qrxy=0.25,
               int _Nt=1,  double _Qt=RADIAN(10),
               int _Ndx=1, double _Qdx=0.1,
               int _Ndy=1, double _Qdy=0.1,
               double _R=1.0, double _THETA=0.0);
        ~GHOUGH();
        
        /* find one contour */
        CONTOUR *localize(CONTOUR *reference, IMAGE *img) ;
        CONTOUR *localize(CONTOUR *reference, EDGE *edgeMap);
  
  	/* find multiple contours */
        CONTOUR **localize(CONTOUR *reference, IMAGE *img, int numFind) ;
        CONTOUR **localize(CONTOUR *reference, EDGE *edgeMap, int numFind);
};

class GSNAKE : public PYRAMID,
	       public CONTOUR  {

    protected :

	double global_lambda;	/* global regularization parameters */	
	double Eint;		/* Total internal energy of snake */
	double Eext;		/* Total external energy of snake */
        double Esnake;          /* Total snake energy */
	EEXTTYPE EextType;      /* external energy type */
	EDGE *EdgeMap;		/* edge map in current pyramid level */
	IMAGE *GaussImg;	/* gaussImg in current pyramid level */

    protected :


	/* dynamic programming */
	int     mn_dymprog( short level, int search_spacing,
		int nhood, unsigned char blowup = 1, 
		int verbose = 0, int showImg = 1, int img_Xoffset = 0, 
		int img_Yoffset = 0);
	double dymprog(int spacing, LOCATIONS *searchLocs, int classify = 0);
	void move( GSNAKE *old_gssnake,
		   int pt_Xoffset = 0, int pt_Yoffset = 0);


	/* regularization. To compute Epoint based on current regularization
	   policy */
	void regularize( SNAXEL *sxptr,
		double *Emodel, double *Eimage, double *Epoint );

	double EEdge( SNAXEL *now, EDGE *edgeMap );
	double EInten( SNAXEL *now, IMAGE *gaussImg );
	double EMag( SNAXEL *now, EDGE *edgeMap );

	double ESnaxel( SNAXEL *sxptr, int pt_ind, int loc_ind, 
		int spacing, LOCATIONS *searchLocs, int verbose = 0 );

    public :

	GSNAKE(EEXTTYPE etype = _EDGE, double glambda=_LOCAL_MINMAX);
	~GSNAKE(void);
	void reset(void);   
	
	/* generate pyramid */
	genPyramid(short level, short verbose) ;

 	/* compute energy routines */
        double ESnake( short level, int verbose = 0 );

	double ESnaxel( SNAXEL *sxptr,
		BOOLEAN store = _FALSE, int verbose = 0 ); 
		

	double EInternal(SNAXEL *now);
       	double EExternal(SNAXEL *now);

	/* localize gsnake routines */
        int localize(int _Qx=1, int _Qy=1, 
               int _NR=1, double  _QR=0.25, 
               int _Nrxy=1, double _Qrxy=0.1,
               int _Nt=1,  double _Qt=RADIAN(10),
               int _Ndx=1, double _Qdx=0.1,
               int _Ndy=1, double _Qdy=0.1,
               double _R=1.0, double _THETA=0.0);
               
	void fineLocalize( int _qx = 3, int _qy = 3, 
			  int _cg_col = 0, int _cg_row = 0 );

	/* minimization routines */
        int     minimize( 
			int segmentSpacing = 5,
			int numSearchSegment = 5,
			int snaxelSpacing  = 5,
			unsigned char blowup = 1,
			int verbose = 0, 
		        double theLambda = _DEFINED_LAMBDA,
		 	int showImg	= 1,
			int img_Xoffset = 0,
			int img_Yoffset = 0 );
	
	/* marginalization routine */
	double marginalize(int nhoodTangent=5, int nhoodNormal=3) ;

	/* interface */
	GSNAKE *duplicate(GSNAKE *target=NULL);
	void deform( unsigned char blowup=1, short expand=1 );

	void showLine( unsigned char blowup=1,
		   int Xoffset = 0, int Yoffset = 0 );
	void show( unsigned char blowup =1,short expand = 1, 
		   int showImg = 1,
		   int img_Xoffset = 0, int img_Yoffset = 0,
		   int pt_Xoffset = 0, int pt_Yoffset = 0
		  );

	double getEint(void) {return Eint;}
	double getEext(void) {return Eext;}
	double getEsnake(void) { return Esnake;}
	EEXTTYPE getEextType( void ) { return EextType; }
	double getGLambda( void ) { return global_lambda; }
	void putGLambda(double _lambda) { global_lambda = _lambda; }

};

class MODEL : public GSNAKE {

    protected :
	short deformSample;	/* # sample of learn deform */
	short shapeSample;	/* # sample of learn shape */

    protected :
	int init( GSNAKE *sample );
	int initShape( GSNAKE *sample );
	int initDeform( GSNAKE * sample );

    public :
	MODEL(void);
	~MODEL(void) {;}
	int LearnShape( GSNAKE *sample );
	int LearnDeform( GSNAKE *sample );
	GSNAKE *getModel(void) { return this; }
};

class CLASSIFY {

   protected :
   	int numClass ;			/* number of class */
   	EEXTTYPE Eexttype ;		/* type of external energy */
   	CONTOUR **templates ;		/* reference templates */
   	char **labels ;
   	double *Margin_Prob;
   	double *Rigid_Match ;
   	double *Deform_Match ;
   	double *Deform_Prob ;
   
   /* these localization and minimization parameters are made public so
      that modifying them can be easy */
   public :
   	int Qx, Qy ;			/* GHT Image Cell Resolutions */
   	int NR, Nrxy, Nt, Ndx, Ndy ;	/* GHT ranges */
   	double QR, Qrxy, Qt, Qdx, Qdy;  /* GHT Resolutions */
   	double R, THETA ;		/* GHT Constants */
   	int numSearchSegment ;		/* number of search segments */
   	int segmentSpacing ;		/* spacing between segments */
   	int numLevel ;			/* number of pyramid levels */
   	int verbose ;			/* verbose mode */
   	int nhoodTangent, nhoodNormal;	/* neighborhood used in marginaliz-n */
   	
   protected :
   	int init(void) ;
   	
   public :
	CLASSIFY(EEXTTYPE _Eexttype = _EDGE);
	~CLASSIFY(void);
	int read( CONTOUR *contour );
	int read( char *filename );
	int classify(IMAGE *testimg, CLASSTYPE ClassType = _MARGIN_PROB) ;
	int getNumClass( void ) { return numClass ; }
	double getScore( int template_id, CLASSTYPE ClassType = _MARGIN_PROB);
	int selectMax( CLASSTYPE ClassType = _MARGIN_PROB);
	char *getLabel(short class_id) ;
	void dump(char *imgName, FILE *stream = stdout) ;
};



#endif
